<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:a="http://www.schema-library.org/add/idrefs"
                xmlns:sic="http://www.schema-library.org/schema-in-context"
                xmlns:h="http://www.schema-library.org/input-schema-and-its-dependency-hierarchy"
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                exclude-result-prefixes="#all" version="2.0">
    
    <!-- 
        This program is a real workhorse. For every reference in a
        schema it adds an idref. Thus, between the program that adds an id to each global item and this program
        I have created an id-idref pair. This makes it easy to jump
        from an item with a reference to the referenced item.
        
        Suppose an element has a type attribute with value A. Suppose A is
        a global complexType. It has an id attribute, say 2DE. This program
        adds an idref attribute whose value is 2DE. Ditto for an attribute
        with a type attribute.
        
        An idref attribute is added to these items:
        
        - elements with a type attribute
        - attributes with a type attribute
        - elements with a ref attribute
        - attributes with a ref attribute
        - group elements with a ref attribute
        - attributeGroup elements with a ref attribute
        - restriction elements (in a simpleType or complexType) with a base attribute
        - extension elements (in a complexType) with a base attribute
        - list elements with an itemType attribute
        - union elements with a memberTypes attribute
        
        This program also adds a targetNamespace attribute onto each
        element and attribute. Obviously, its value is the namespace
        of the item. What about a no-namespace schema that is coerced
        into the namespace of an including schema? No problem. My program
        handles that.
        
        Finally, if an element substitutes for another element, I add
        an attribute, substitutionGroupRef, to the former.
        
        Author: Roger L. Costello
    -->
    
    <xsl:function name="a:add-idref-substitutionGroupRef-and-targetNamespace-to-items-in-all-schemas" as="document-node()">
        <xsl:param name="schemas" as="document-node()" />
        
        <xsl:variable name="schema" select="$schemas/schemas/xs:schema[1]" />
        <xsl:variable name="schema-in-context" select="sic:schema-in-context($schemas/schemas/xs:schema[1])" />
        
        <xsl:variable name="updated-schemas">
            <xsl:sequence select="a:add-idref-substitutionGroupRef-and-targetNamespace-to-items-in-all-schemas($schema, $schema-in-context)" />
        </xsl:variable>
        
        <xsl:sequence select="$updated-schemas" />
        
    </xsl:function>
    
    <xsl:function name="a:add-idref-substitutionGroupRef-and-targetNamespace-to-items-in-all-schemas" as="document-node()">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:variable name="schema-item" as="element()*">
            <container>
                <item><xsl:sequence select="$schema" /></item>
                <schema-in-context><xsl:sequence select="$schema-in-context" /></schema-in-context>
            </container>
        </xsl:variable>
        
        <xsl:variable name="updated-schemas">
            <schemas> 
                <xsl:apply-templates select="$process-graph">
                    <xsl:with-param name="items-to-process" select="$schema-item" />
                    <xsl:with-param name="items-already-processed-or-in-queue-for-processing" select="$schema-item" />
                    <xsl:with-param name="function-to-process-current-item" select="$a:function-to-process-current-schema" />
                    <xsl:with-param name="function-to-determine-new-items-to-process" select="$a:function-to-determine-new-schemas-to-process" />
                </xsl:apply-templates> 
            </schemas>  
        </xsl:variable>
        
        <xsl:sequence select="$updated-schemas" />
        
    </xsl:function>
    
    
    <xsl:variable name="a:function-to-determine-new-schemas-to-process" select="document('')//a:function-to-determine-new-schemas-to-process[1]" />
    <xsl:variable name="a:function-to-process-current-schema" select="document('')//a:function-to-process-current-schema[1]" />
    
    <a:function-to-determine-new-schemas-to-process />
    <a:function-to-process-current-schema />
    
    
    <!-- 
        The new schemas are those pointed at by xs:include, xs:import, and xs:redefine which are not already on
        the list for processing.
    -->
    
    <xsl:template match="a:function-to-determine-new-schemas-to-process">
        <xsl:param name="item" />
        <xsl:param name="items-already-processed-or-in-queue-for-processing" />
         
        <xsl:variable name="schema-in-context" select="$item/schema-in-context/*" />
        <xsl:variable name="items" select="$item/item/xs:schema/(xs:include[@schemaLocation] | xs:import[@schemaLocation] | xs:redefine[@schemaLocation])" />
        <xsl:sequence select="a:helper-function-to-determine-new-schemas-to-process($schema-in-context, $items, $items-already-processed-or-in-queue-for-processing)" />
        
    </xsl:template>
    
    <xsl:function name="a:helper-function-to-determine-new-schemas-to-process">
        <xsl:param name="schema-in-context" />
        <xsl:param name="items" />
        <xsl:param name="items-already-processed-or-in-queue-for-processing" />
        
        <xsl:choose>
            <xsl:when test="empty($items)" />
            <xsl:otherwise>
                <xsl:for-each select="$items[1]">
                    
                    <xsl:variable name="i" select="." />
                    <xsl:variable name="referenced-schema" select="$schemas/schemas/xs:schema[@_id eq $i/@idref]" />
                    
                    <xsl:choose>
                        <xsl:when test="not($referenced-schema/@_id = $items-already-processed-or-in-queue-for-processing/item/xs:schema/@_id)">
                            <xsl:variable name="schema-item" as="element()*">
                                <container>
                                    <item><xsl:sequence select="$referenced-schema" /></item>
                                    <schema-in-context><xsl:sequence select="sic:schema-in-context($referenced-schema, $schema-in-context)" /></schema-in-context>
                                </container>
                            </xsl:variable>
                            <xsl:sequence select="$schema-item" />
                            <xsl:sequence select="a:helper-function-to-determine-new-schemas-to-process($schema-in-context, $items[position() gt 1], ($items-already-processed-or-in-queue-for-processing, $schema-item))" />
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:sequence select="a:helper-function-to-determine-new-schemas-to-process($schema-in-context, $items[position() gt 1], $items-already-processed-or-in-queue-for-processing)" />
                        </xsl:otherwise>
                    </xsl:choose>
                    
                </xsl:for-each>
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    

    <xsl:template match="a:function-to-process-current-schema">
        <xsl:param name="item" />
        
        <xsl:variable name="schema" select="$item/item/*" />
        <xsl:variable name="schema-in-context" select="$item/schema-in-context/*" />
        
        <xsl:sequence select="a:add-idref-substitutionGroupRef-and-targetNamespace-to-schema-items($schema, $schema-in-context)" />
        
    </xsl:template> 
    

    <xsl:function name="a:add-idref-substitutionGroupRef-and-targetNamespace-to-schema-items">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:apply-templates select="$schema" mode="add-idref-substitutionGroupRef-and-targetNamespace">
            <xsl:with-param name="schema" select="$schema" />
            <xsl:with-param name="schema-in-context" select="$schema-in-context" />
        </xsl:apply-templates>
        
     </xsl:function>
    
    <xsl:template match="/" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:copy>
            <xsl:apply-templates select="node()"  mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="@* | node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:copy>
            <xsl:apply-templates select="@*"  mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
            <xsl:apply-templates  mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="xs:element[@substitutionGroup][not(@type or xs:simpleType or xs:complexType)]" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:variable name="id-of-element-being-substituted" select="a:get-id-of-referenced-item(./@substitutionGroup, ., $schema-in-context, ('element'))" />
        
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:attribute name="substitutionGroupRef"><xsl:value-of select="$id-of-element-being-substituted" /></xsl:attribute>
            <xsl:attribute name="targetNamespace"><xsl:value-of select="sic:namespace($schema-in-context)" /></xsl:attribute>
            <xsl:apply-templates select="node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="xs:element[@substitutionGroup][@type]" mode="add-idref-substitutionGroupRef-and-targetNamespace" priority="99">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:variable name="id-of-element-being-substituted" select="a:get-id-of-referenced-item(./@substitutionGroup, ., $schema-in-context, ('element'))" />
        
        <xsl:variable name="id-of-referenced-type" select="a:get-id-of-referenced-item(./@type, ., $schema-in-context, ('simpleType', 'complexType'))" />
        
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:attribute name="substitutionGroupRef"><xsl:value-of select="$id-of-element-being-substituted" /></xsl:attribute>
            <xsl:attribute name="idref"><xsl:value-of select="$id-of-referenced-type" /></xsl:attribute>
            <xsl:attribute name="targetNamespace"><xsl:value-of select="sic:namespace($schema-in-context)" /></xsl:attribute>
            <xsl:apply-templates select="node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="xs:element[@type] | xs:attribute[@type]" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:variable name="id-of-referenced-type" select="a:get-id-of-referenced-item(./@type, ., $schema-in-context, ('simpleType', 'complexType'))" />
        
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:attribute name="idref"><xsl:value-of select="$id-of-referenced-type" /></xsl:attribute>
            <xsl:attribute name="targetNamespace"><xsl:value-of select="sic:namespace($schema-in-context)" /></xsl:attribute>
            <xsl:apply-templates select="node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="xs:element[@ref] | xs:attribute[@ref]" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:variable name="id-of-referenced-item" select="a:get-id-of-referenced-item(./@ref, ., $schema-in-context, ('element', 'attribute'))" />
        <xsl:variable name="ns-of-referenced-item" select="a:get-ns-of-referenced-item(./@ref, ., $schema-in-context, ('element', 'attribute'))" />
        
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:attribute name="idref"><xsl:value-of select="$id-of-referenced-item" /></xsl:attribute>
            <xsl:attribute name="targetNamespace"><xsl:value-of select="$ns-of-referenced-item" /></xsl:attribute>
            <xsl:apply-templates select="node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="xs:element | xs:attribute" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:attribute name="targetNamespace"><xsl:value-of select="sic:namespace($schema-in-context)" /></xsl:attribute>
            <xsl:apply-templates select="node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="xs:group[@ref] | xs:attributeGroup[@ref]" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:variable name="id-of-referenced-group" select="a:get-id-of-referenced-item(./@ref, ., $schema-in-context, ('group', 'attributeGroup'))" />
        
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:attribute name="idref"><xsl:value-of select="$id-of-referenced-group" /></xsl:attribute>
            <xsl:apply-templates select="node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="xs:restriction[@base] | xs:extension[@base]" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:variable name="id-of-referenced-item" select="a:get-id-of-referenced-item(./@base, ., $schema-in-context, ('simpleType', 'complexType'))" />
        
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:attribute name="idref"><xsl:value-of select="$id-of-referenced-item" /></xsl:attribute>
            <xsl:apply-templates select="node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="xs:list[@itemType]" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:variable name="id-of-referenced-item" select="a:get-id-of-referenced-item(./@itemType, ., $schema-in-context, ('simpleType'))" />
        
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:attribute name="idref"><xsl:value-of select="$id-of-referenced-item" /></xsl:attribute>
            <xsl:apply-templates select="node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    <xsl:template match="xs:union[@memberTypes]" mode="add-idref-substitutionGroupRef-and-targetNamespace">
        <xsl:param name="schema"/>
        <xsl:param name="schema-in-context"/>
        
        <xsl:variable name="schema-namespace" select="sic:namespace($schema-in-context)" />
        <xsl:variable name="union" select="." />
        
        <xsl:variable name="ids-of-referenced-items">
            <xsl:for-each select="tokenize(normalize-space(@memberTypes), '\s+')">
                <xsl:variable name="QName" select="resolve-QName(., $union)"/>
                <xsl:variable name="name" select="local-name-from-QName($QName)"/>
                <xsl:variable name="ns-tmp" select="namespace-uri-from-QName($QName)" />
                <xsl:variable name="ns" select="if (($ns-tmp eq '') and ($schema-namespace ne '')) then $schema-namespace else $ns-tmp" />
                
                <xsl:variable name="in-scope-schemas" select="sic:in-scope-schemas-with-namespace($schema-in-context, $ns)"/>
                <xsl:variable name="refed-simpleType" select="$in-scope-schemas/xs:simpleType[@name eq $name]"/>
                
                <xsl:value-of select="$refed-simpleType/@_id" />
                <xsl:text> </xsl:text>
            </xsl:for-each>
        </xsl:variable>
        
        <xsl:copy>
            <xsl:copy-of select="@*" />
            <xsl:attribute name="idrefs"><xsl:value-of select="normalize-space($ids-of-referenced-items)" /></xsl:attribute>
            <xsl:apply-templates select="node()" mode="add-idref-substitutionGroupRef-and-targetNamespace">
                <xsl:with-param name="schema" select="$schema" />
                <xsl:with-param name="schema-in-context" select="$schema-in-context" />
            </xsl:apply-templates>
        </xsl:copy>
        
    </xsl:template>
    
    
    <xsl:function name="a:get-id-of-referenced-item">
        <xsl:param name="reference" />
        <xsl:param name="item" />
        <xsl:param name="schema-in-context" />
        <xsl:param name="kinds-of-referenced-items" />

        <xsl:variable name="ns" select="a:get-ns-of-referenced-item($reference, $item, $schema-in-context, $kinds-of-referenced-items)" />
        
        <xsl:variable name="in-scope-schemas" select="sic:in-scope-schemas-with-namespace($schema-in-context, $ns)" />

        <xsl:variable name="QName" select="resolve-QName($reference, $item)" />
        <xsl:variable name="name" select="local-name-from-QName($QName)" />
        
        <xsl:variable name="referenced-item" select="$in-scope-schemas/descendant-or-self::xs:schema/xs:*[local-name() = $kinds-of-referenced-items][@name eq $name]" />
        
        <xsl:if test="empty($referenced-item)">
            <xsl:message terminate="yes">
                <xsl:text>[ ERROR IN a:get-id-of-referenced-item ]</xsl:text>
                <in-scope-schemas><xsl:sequence select="$in-scope-schemas" /></in-scope-schemas>
                <item><xsl:sequence select="$item" /></item>
                <ns><xsl:value-of select="$ns" /></ns>
                <name><xsl:value-of select="$name" /></name>
                <kinds-of-referenced-items><xsl:value-of select="$kinds-of-referenced-items" /></kinds-of-referenced-items>
            </xsl:message>
        </xsl:if>
        
        <xsl:value-of select="$referenced-item/@_id" />
        
    </xsl:function>
    
    
    <xsl:function name="a:get-ns-of-referenced-item">
        <xsl:param name="reference" />
        <xsl:param name="item" />
        <xsl:param name="schema-in-context" />
        <xsl:param name="kinds-of-referenced-items" />
        
        <xsl:variable name="schema-namespace" select="sic:namespace($schema-in-context)" />
        
        <xsl:variable name="namespace-of-schema-in-isolation" select="sic:namespace(sic:schema-without-context($schema-in-context))" />
        
        <xsl:variable name="QName" select="resolve-QName($reference, $item)" />
        <xsl:variable name="name" select="local-name-from-QName($QName)" />
        <xsl:variable name="ns-tmp" select="namespace-uri-from-QName($QName)" />
        
        <!-- Consider a reference to a component in a no-namespace schema.
             If the no-namespace schema is xs:included then the component is 
             namespace-coerced and the referencing element is
             in the namespace of the schema it is located in. If the 
             no-namespace schema is xs:imported then the component is not 
             namespace-coerced and the referencing element is in no-namespace.
             
             For the following cases assume that the reference has no namespace ($ns-tmp eq '').
             Case 1: The reference is in a schema (A) with no-namespace and A is included by a 
             schema (B) with a namespace. Then the in-scope-schemas are all those schemas with
             namespace(B).
             Case 2: The reference is in a schema (A) with a namespace and the reference is to
             an *imported* schema (B) that has no namespace. Then the in-scope-schemas are all those
             schemas with namespace("").
             Conclusion: if namespace-of-schema-in-isolation(A) eq '' and schema-namespace ne '' then
                              the in-scope-schemas are all those schemas with namespace(B)
                         else if namespace-of-schema-in-isolation(A) ne '' then
                              the in-scope-schemas are all those schemas with namespace("")
                         else 
                              the in-scope-schemas are all those schemas with namespace("")
        -->
        
        <!--xsl:variable name="ns" select="if (($ns-tmp eq '') and ($namespace-of-schema-in-isolation eq '') and ($schema-namespace ne '')) then $schema-namespace else $ns-tmp" /-->

        <xsl:variable name="ns" select="if ($ns-tmp eq '') then                                   (: The reference is to an item in no-namespace :)
                                              if ($namespace-of-schema-in-isolation eq '') then      (: The reference is in a schema with no-namespace :)
                                                    if ($schema-namespace ne '') then                    (: The no-namespace schema is being namespace-coerced :) 
                                                           $schema-namespace                                 (: Return the namespace that the schema is being coerced to :)
                                                    else                                                 (: The no-namespace schema is not being namespace-coerced:)
                                                           $ns-tmp                                           (: Return no-namespace :)
                                              else                                                   (: The reference is in a schema with a namespace - the referenced item must be in an imported schema :)
                                                    $ns-tmp                                              (: Return no-namespace :)
                                         else                                                     (: The reference is to an item in a namespace :)
                                              $ns-tmp" />                                             <!-- Return namespace -->
        
        <xsl:value-of select="$ns" />
        
    </xsl:function>
     
</xsl:stylesheet>
