<?xml version="1.0"?>
<xsl:transform xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
               xmlns:xs="http://www.w3.org/2001/XMLSchema"
               xmlns:_="http://www.functions.org/higher-order/" 
               version="2.0">
    
    <!--
        Author: Roger L. Costello
        Date: February 2011 
    -->   
    
    
    <xsl:variable name="until" select="document('')//_:until[1]" />
    
    <xsl:variable name="process-graph" select="document('')//_:process-graph[1]" />
    
    <xsl:variable name="map" select="document('')//_:map[1]" />
    <xsl:variable name="filter" select="document('')//_:filter[1]" />
    <xsl:variable name="foldl" select="document('')//_:foldl[1]" />
    <xsl:variable name="foldr" select="document('')//_:foldr[1]" />
    
    <_:until />
    
    <_:process-graph />
    
    <_:map />
    <_:filter />
    <_:foldl />
    <_:foldr />
    
    <!-- 
        until :: (a -> Bool) -> (a -> a) -> a -> a
        until p f x = if p x then x else until p f (f x)
        
        p      = a function that has the condition for when to stop processing
        f      = a function that is applied to the argument, x 
        x      = the value being operated upon
    -->
    <xsl:template match="_:until">
        <xsl:param name="p" />
        <xsl:param name="f" />
        <xsl:param name="x" />
          
        <xsl:variable name="finished" as="xs:boolean">
            <xsl:apply-templates select="$p">
                <xsl:with-param name="item" select="$x" />
            </xsl:apply-templates>
        </xsl:variable>
        <xsl:message>$finished = <xsl:value-of select="$finished" /></xsl:message>
        <xsl:choose>
            <xsl:when test="$finished">
                <xsl:sequence select="$x" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:variable name="x_">
                    <xsl:apply-templates select="$f">
                        <xsl:with-param name="item" select="$x" />
                    </xsl:apply-templates>
                </xsl:variable>
                <xsl:apply-templates select="$until">
                    <xsl:with-param name="p" select="$p" />
                    <xsl:with-param name="f" select="$f" />
                    <xsl:with-param name="x" select="$x_" />
                </xsl:apply-templates>
                
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:template>
    
    <!-- 
        process-graph [] _ _ _ = []
        process-graph (x:xs) ys f g = (f x) ++ (process-graph (xs ++ g(x)) (ys ++ g(x)) f g)
        
        (x:xs) = items-to-process, x is the first item and xs are the remaining items
        ys     = items-already-processed-or-in-queue-for-processing
        f      = function-to-process-current-item
        g      = function-to-determine-new-items-to-process
    -->
    <xsl:template match="_:process-graph">
        <xsl:param name="items-to-process" />
        <xsl:param name="items-already-processed-or-in-queue-for-processing" />
        <xsl:param name="function-to-process-current-item" />
        <xsl:param name="function-to-determine-new-items-to-process" />
        
        <xsl:choose>
            <xsl:when test="empty($items-to-process)" />
            <xsl:otherwise>
                
                <xsl:variable name="item" select="$items-to-process[1]" />
                
                <xsl:variable name="processed-item" as="item()*">
                    <xsl:apply-templates select="$function-to-process-current-item">
                        <xsl:with-param name="item" select="$item" />
                    </xsl:apply-templates>
                </xsl:variable>
                
                <xsl:sequence select="$processed-item" />
                
                <xsl:variable name="new-items-to-process" as="item()*">
                    <xsl:apply-templates select="$function-to-determine-new-items-to-process">
                        <xsl:with-param name="item" select="$item" />
                        <xsl:with-param name="items-already-processed-or-in-queue-for-processing" select="$items-already-processed-or-in-queue-for-processing" />
                    </xsl:apply-templates>
                </xsl:variable>
                
                <xsl:apply-templates select="$process-graph">
                    <xsl:with-param name="items-to-process" select="$items-to-process[position() gt 1] | $new-items-to-process" />
                    <xsl:with-param name="items-already-processed-or-in-queue-for-processing" select="$items-already-processed-or-in-queue-for-processing | $new-items-to-process" />
                    <xsl:with-param name="function-to-process-current-item" select="$function-to-process-current-item" />
                    <xsl:with-param name="function-to-determine-new-items-to-process" select="$function-to-determine-new-items-to-process" />
                </xsl:apply-templates>
                
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:template>
    
    <!-- 
        This map function creates a list that is the same length as "list"
        where each item in the new list is the result of applying "function"
        to the corresponding item in "list".
    -->
    
    <xsl:template match="_:map">
        <xsl:param name="function" />
        <xsl:param name="list" />
        
        <xsl:choose>
            <xsl:when test="not($list)" />
            <xsl:when test="not($function)" />
            <xsl:otherwise>
                <xsl:apply-templates select="$function">
                    <xsl:with-param name="item" select="$list[1]" />
                </xsl:apply-templates>
                <xsl:apply-templates select="$map">
                    <xsl:with-param name="function" select="$function" />
                    <xsl:with-param name="list" select="$list[position() &gt; 1]" />
                </xsl:apply-templates>
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:template>
    
    <!-- 
        The "function" parameter must return a Boolean value (true or false) 
        This filter function creates a list consisting of all the items in
        "list" which evaluate to true by "function".
    -->
    
    <xsl:template match="_:filter">
        <xsl:param name="function" />
        <xsl:param name="list" />
        
        <xsl:choose>
            <xsl:when test="not($list)" />
            <xsl:when test="not($function)" />
            <xsl:otherwise>
                <xsl:variable name="boolean">
                    <xsl:apply-templates select="$function">
                        <xsl:with-param name="item" select="$list[1]" />
                    </xsl:apply-templates>
                </xsl:variable>
                <xsl:choose>
                    <xsl:when test="string($boolean) = 'true'">
                        <xsl:copy-of select="$list[1]"/>
                    </xsl:when>
                </xsl:choose>
                <xsl:apply-templates select="$filter">
                    <xsl:with-param name="function" select="$function" />
                    <xsl:with-param name="list" select="$list[position() &gt; 1]" />
                </xsl:apply-templates>
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:template>
    
    <xsl:template match="_:foldl">
        <xsl:param name="function" />
        <xsl:param name="accumulator" />
        <xsl:param name="list" />
        
        <xsl:choose>
            <xsl:when test="not($list)">
                <xsl:copy-of select="$accumulator" />
            </xsl:when>
            <xsl:when test="not($function)">
                <xsl:copy-of select="$accumulator" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:variable name="updated-accumulator">
                    <xsl:apply-templates select="$function">
                        <xsl:with-param name="accumulator" select="$accumulator" />
                        <xsl:with-param name="item" select="$list[1]" />
                    </xsl:apply-templates>
                </xsl:variable>
                <xsl:apply-templates select="$foldl">
                    <xsl:with-param name="function" select="$function" />
                    <xsl:with-param name="accumulator" select="$updated-accumulator" />
                    <xsl:with-param name="list" select="$list[position() &gt; 1]" />
                </xsl:apply-templates>
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:template>
    
    <xsl:template match="_:foldr">
        <xsl:param name="function" />
        <xsl:param name="accumulator" />
        <xsl:param name="list" />
        
        <xsl:choose>
            <xsl:when test="not($list)">
                <xsl:copy-of select="$accumulator" />
            </xsl:when>
            <xsl:when test="not($function)">
                <xsl:copy-of select="$accumulator" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:variable name="updated-accumulator">
                    <xsl:apply-templates select="$foldr">
                        <xsl:with-param name="function" select="$function" />
                        <xsl:with-param name="accumulator" select="$accumulator" />
                        <xsl:with-param name="list" select="$list[position() &gt; 1]" />
                    </xsl:apply-templates>
                </xsl:variable>
                <xsl:apply-templates select="$function">
                    <xsl:with-param name="accumulator" select="$updated-accumulator" />
                    <xsl:with-param name="item" select="$list[1]" />
                </xsl:apply-templates>
            </xsl:otherwise>
        </xsl:choose>
        
    </xsl:template>
    
</xsl:transform>